import asyncio
import os
from typing import Awaitable, Callable, Optional, Union

from fastapi import Request, WebSocket
from flet.core.page import Page
from flet.core.types import WebRenderer

from flet_web.fastapi.flet_app import (
    DEFAULT_FLET_OAUTH_STATE_TIMEOUT,
    DEFAULT_FLET_SESSION_TIMEOUT,
    FletApp,
)
from flet_web.fastapi.flet_fastapi import FastAPI
from flet_web.fastapi.flet_oauth import FletOAuth
from flet_web.fastapi.flet_static_files import FletStaticFiles
from flet_web.fastapi.flet_upload import FletUpload


def app(
    session_handler: Union[Callable[[Page], Awaitable], Callable[[Page], None]],
    proxy_path: Optional[str] = None,
    assets_dir: Optional[str] = None,
    app_name: Optional[str] = None,
    app_short_name: Optional[str] = None,
    app_description: Optional[str] = None,
    web_renderer: WebRenderer = WebRenderer.CANVAS_KIT,
    use_color_emoji: bool = False,
    route_url_strategy: str = "path",
    upload_dir: Optional[str] = None,
    upload_endpoint_path: Optional[str] = None,
    max_upload_size: Optional[int] = None,
    secret_key: Optional[str] = None,
    session_timeout_seconds: int = DEFAULT_FLET_SESSION_TIMEOUT,
    oauth_state_timeout_seconds: int = DEFAULT_FLET_OAUTH_STATE_TIMEOUT,
):
    """
    Mount all Flet FastAPI handlers in one call.

    Parameters:
    * `session_handler` (function or coroutine) - application entry point - a method called for newly connected user. Handler must have 1 parameter: `page` - `Page` instance.
    * `assets_dir` (str, optional) - an absolute path to app's assets directory.
    * `app_name` (str, optional) - PWA application name.
    * `app_short_name` (str, optional) - PWA application short name.
    * `app_description` (str, optional) - PWA application description.
    * `web_renderer` (WebRenderer) - web renderer defaulting to `WebRenderer.CANVAS_KIT`.
    * `use_color_emoji` (bool) - whether to load a font with color emoji. Default is `False`.
    * `route_url_strategy` (str) - routing URL strategy: `path` (default) or `hash`.
    * `upload_dir` (str) - an absolute path to a directory with uploaded files.
    * `upload_endpoint_path` (str, optional) - absolute URL of upload endpoint, e.g. `/upload`.
    * `max_upload_size` (str, int) - maximum size of a single upload, bytes. Unlimited if `None`.
    * `secret_key` (str, optional) - secret key to sign and verify upload requests.
    * `session_timeout_seconds` (int, optional)- session lifetime, in seconds, after user disconnected.
    * `oauth_state_timeout_seconds` (int, optional) - OAuth state lifetime, in seconds, which is a maximum allowed time between starting OAuth flow and redirecting to OAuth callback URL.
    """

    env_upload_dir = os.getenv("FLET_UPLOAD_DIR")
    if env_upload_dir:
        upload_dir = env_upload_dir

    env_websocket_endpoint = os.getenv("FLET_WEBSOCKET_HANDLER_ENDPOINT")
    websocket_endpoint = (
        "ws" if not env_websocket_endpoint else env_websocket_endpoint.strip("/")
    )

    env_upload_endpoint = os.getenv("FLET_UPLOAD_HANDLER_ENDPOINT")
    upload_endpoint = (
        "upload" if not env_upload_endpoint else env_upload_endpoint.strip("/")
    )

    env_oauth_callback_endpoint = os.getenv("FLET_OAUTH_CALLBACK_HANDLER_ENDPOINT")
    oauth_callback_endpoint = (
        "oauth_callback"
        if not env_oauth_callback_endpoint
        else env_oauth_callback_endpoint.strip("/")
    )

    fastapi_app = FastAPI()

    @fastapi_app.websocket(f"/{websocket_endpoint}")
    async def app_handler(websocket: WebSocket):
        await FletApp(
            asyncio.get_running_loop(),
            session_handler,
            session_timeout_seconds=session_timeout_seconds,
            oauth_state_timeout_seconds=oauth_state_timeout_seconds,
            upload_endpoint_path=upload_endpoint_path,
            secret_key=secret_key,
        ).handle(websocket)

    if upload_dir:

        @fastapi_app.put(
            f"/{upload_endpoint_path if upload_endpoint_path else upload_endpoint}"
        )
        async def upload_handler(request: Request):
            await FletUpload(
                upload_dir=upload_dir,
                max_upload_size=max_upload_size,
                secret_key=secret_key,
            ).handle(request)

    @fastapi_app.get(f"/{oauth_callback_endpoint}")
    async def oauth_redirect_handler(request: Request):
        return await FletOAuth().handle(request)

    fastapi_app.mount(
        path="/",
        app=FletStaticFiles(
            proxy_path=proxy_path,
            assets_dir=assets_dir,
            app_name=app_name,
            app_short_name=app_short_name,
            app_description=app_description,
            web_renderer=web_renderer,
            use_color_emoji=use_color_emoji,
            route_url_strategy=route_url_strategy,
        ),
    )

    return fastapi_app



import os
from pathlib import Path

from flet_web.patch_index import patch_index_html, patch_manifest_json


def get_package_web_dir():
    web_root_dir = os.environ.get("FLET_WEB_PATH")
    return web_root_dir or str(Path(__file__).parent.joinpath("web"))



import dataclasses
from datetime import datetime
from typing import Optional


@dataclasses.dataclass
class OAuthState:
    session_id: str
    expires_at: datetime
    complete_page_url: Optional[str] = dataclasses.field(default=None)
    complete_page_html: Optional[str] = dataclasses.field(default=None)



import logging
import os
from datetime import datetime, timezone
from typing import Optional

import flet_web.fastapi as flet_fastapi
from anyio import open_file
from fastapi import Request
from flet_web.uploads import build_upload_query_string, get_upload_signature

logger = logging.getLogger(flet_fastapi.__name__)


class FletUpload:
    """
    Flet app uploads handler.

    Parameters:

    * `upload_dir` (str) - an absolute path to a directory with uploaded files.
    * `max_upload_size` (str, int) - maximum size of a single upload, bytes. Unlimited if `None`.
    * `secret_key` (str, optional) - secret key to sign and verify upload requests.
    """

    def __init__(
        self,
        upload_dir: str,
        max_upload_size: Optional[int] = None,
        secret_key: Optional[str] = None,
    ) -> None:
        self.__upload_dir = os.path.realpath(upload_dir)
        self.__max_upload_size = max_upload_size

        env_max_upload_size = os.getenv("FLET_MAX_UPLOAD_SIZE")
        if env_max_upload_size:
            self.__max_upload_size = int(env_max_upload_size)

        self.__secret_key = secret_key

        env_upload_secret_key = os.getenv("FLET_SECRET_KEY")
        if env_upload_secret_key:
            self.__secret_key = env_upload_secret_key

        logger.info(f"Upload path configured: {self.__upload_dir}")

    """
    Handle file upload.

    Upload must be an non-encoded (raw) file in the requst body.
    """

    async def handle(self, request: Request):
        file_name = request.query_params["f"]
        expire_str = request.query_params["e"]
        signature = request.query_params["s"]

        if not file_name or not expire_str or not signature:
            raise Exception("Invalid request")

        expire_date = datetime.fromisoformat(expire_str)

        # verify signature
        query_string = build_upload_query_string(file_name, expire_date)
        if (
            get_upload_signature(
                request.url.path, query_string, expire_date, self.__secret_key
            )
            != signature
        ):
            raise Exception("Invalid request")

        # check expiration date
        if datetime.now(timezone.utc) >= expire_date:
            raise Exception("Invalid request")

        # build/validate dest path
        joined_path = os.path.join(self.__upload_dir, file_name)
        full_path = os.path.realpath(joined_path)
        if os.path.commonpath([full_path, self.__upload_dir]) != self.__upload_dir:
            raise Exception("Invalid request")

        # create directory if not exists
        dest_dir = os.path.dirname(full_path)
        os.makedirs(dest_dir, exist_ok=True)

        # upload file
        size = 0
        async with await open_file(full_path, "wb") as f:
            async for chunk in request.stream():
                size += len(chunk)
                if self.__max_upload_size and size > self.__max_upload_size:
                    raise Exception(
                        f"Max upload size reached: {self.__max_upload_size}"
                    )
                await f.write(chunk)



import hashlib
import hmac
import logging
import os
import urllib.parse
from datetime import datetime, timedelta, timezone
from typing import Optional

import flet_web

logger = logging.getLogger(flet_web.__name__)


def build_upload_url(
    upload_endpoint_path: str,
    file_name: str,
    expires_in_seconds: int,
    secret_key: Optional[str],
):
    expire_date = datetime.now(timezone.utc) + timedelta(seconds=expires_in_seconds)
    query_string = build_upload_query_string(file_name, expire_date)
    signature = get_upload_signature(
        upload_endpoint_path, query_string, expire_date, secret_key
    )
    return f"/{upload_endpoint_path.strip('/')}?{query_string}&s={signature}"


def build_upload_query_string(file_name: str, expire_date: datetime):
    return urllib.parse.urlencode({"f": file_name, "e": expire_date.isoformat()})


def get_upload_signature(
    upload_endpoint_path: str,
    query_string: str,
    expire_date: datetime,
    secret_key: Optional[str],
):
    env_secret_key = os.getenv("FLET_SECRET_KEY")
    if env_secret_key:
        secret_key = env_secret_key
    if not secret_key:
        raise Exception(
            "Specify secret_key parameter or set FLET_SECRET_KEY environment variable to enable uploads."
        )
    signing_key = hmac.new(
        secret_key.encode("utf-8"),
        expire_date.isoformat().encode("utf-8"),
        hashlib.sha256,
    ).digest()
    return hmac.new(
        signing_key,
        f"{upload_endpoint_path.strip('/')}{query_string}".encode("utf-8"),
        hashlib.sha256,
    ).hexdigest()
